import base64
from typing import Any

from django.db import models
import dacite
from ptz_app.functions.camera import Settings


class CameraSettings(models.Model):
    """Default camera settings."""

    iris = models.IntegerField(null=True)
    shutter = models.IntegerField(null=True)
    gain = models.IntegerField(null=True)
    drc = models.IntegerField(null=True)
    red_gain = models.IntegerField(null=True)
    blue_gain = models.IntegerField(null=True)
    saturation = models.IntegerField(null=True)
    hue = models.IntegerField(null=True)
    brightness = models.IntegerField(null=True)
    contrast = models.IntegerField(null=True)
    sharpness = models.IntegerField(null=True)
    gamma = models.IntegerField(null=True)
    color_temperature = models.IntegerField(null=True)
    noise2d = models.IntegerField(null=True)
    noise3d = models.IntegerField(null=True)

    def json(self) -> dict[str, Any]:
        """Return a JSON representation of the camera settings."""
        return {
            "id": self.id,
            "iris": self.iris,
            "shutter": self.shutter,
            "gain": self.gain,
            "drc": self.drc,
            "red_gain": self.red_gain,
            "blue_gain": self.blue_gain,
            "saturation": self.saturation,
            "hue": self.hue,
            "brightness": self.brightness,
            "contrast": self.contrast,
            "sharpness": self.sharpness,
            "gamma": self.gamma,
            "color_temperature": self.color_temperature,
            "noise2d": self.noise2d,
            "noise3d": self.noise3d,
        }

    def to_dataclass(self, default_settings: "CameraSettings | None" = None) -> Settings:
        """Return a dataclass representation of the camera settings."""
        json_settings = self.json()
        del json_settings["id"]
        if default_settings is not None:
            default_json = default_settings.json()
            del default_json["id"]
            for key in default_json:
                if json_settings[key] is None:
                    json_settings[key] = default_json[key]
        return dacite.from_dict(Settings, json_settings)


class Camera(models.Model):
    """Camera model."""

    name = models.CharField(max_length=100)
    ip = models.GenericIPAddressField()
    username = models.CharField(max_length=100, null=True)
    password = models.CharField(max_length=100, null=True)
    default_settings = models.ForeignKey(CameraSettings, on_delete=models.CASCADE)

    def json(self) -> dict[str, Any]:
        """Return a JSON representation of the camera."""
        return {
            "id": self.id,
            "name": self.name,
            "ip": self.ip,
            "username": self.username,
            "password": self.password,
            "default_settings": self.default_settings.json(),
            "presets": [preset.json() for preset in self.preset_set.all().order_by("order")],
        }


class Preset(models.Model):
    """Camera preset model."""

    name = models.CharField(max_length=100)
    order = models.IntegerField()
    thumbnail = models.BinaryField()
    camera = models.ForeignKey(Camera, on_delete=models.CASCADE)
    settings = models.ForeignKey(CameraSettings, on_delete=models.CASCADE)
    pan = models.IntegerField()
    tilt = models.IntegerField()
    zoom = models.IntegerField()
    focus = models.IntegerField()

    def json(self) -> dict[str, Any]:
        """Return a JSON representation of the camera preset."""
        return {
            "id": self.id,
            "name": self.name,
            "order": self.order,
            "thumbnail": "data:image/jpeg;base64," + base64.b64encode(self.thumbnail).decode(),
            "settings": self.settings.json(),
            "pan": self.pan,
            "tilt": self.tilt,
            "zoom": self.zoom,
            "focus": self.focus,
        }



# Generated by Django 5.0.3 on 2024-03-18 21:17

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="CameraSettings",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("aperture", models.FloatField(null=True)),
                ("brightness", models.IntegerField(null=True)),
                ("saturation", models.IntegerField(null=True)),
                ("contrast", models.IntegerField(null=True)),
                ("sharpness", models.IntegerField(null=True)),
                ("hue", models.IntegerField(null=True)),
            ],
        ),
        migrations.CreateModel(
            name="Camera",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("ip", models.GenericIPAddressField()),
                ("username", models.CharField(max_length=100, null=True)),
                ("password", models.CharField(max_length=100, null=True)),
                (
                    "default_settings",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="ptz_app.camerasettings",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Preset",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("order", models.IntegerField()),
                ("thumbnail", models.ImageField(upload_to="")),
                ("pan", models.FloatField()),
                ("tilt", models.FloatField()),
                ("zoom", models.FloatField()),
                ("focus", models.FloatField()),
                (
                    "camera",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="ptz_app.camera"),
                ),
                (
                    "settings",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="ptz_app.camerasettings",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 5.0.3 on 2024-03-23 01:01

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("ptz_app", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="preset",
            name="thumbnail",
            field=models.BinaryField(),
        ),
    ]



"""Django views for the PTZ app."""

from ptz_app.views.index import index
from ptz_app.views.cameras import (
    new_camera,
    edit_camera,
    upsert_camera,
    delete_camera,
    get_camera_controls,
    update_camera_controls,
    preview_settings,
)
from ptz_app.views.presets import (
    new_preset,
    edit_preset,
    upsert_preset,
    delete_preset,
    recall_preset,
    update_thumbnail,
)

__all__ = [
    "index",
    "new_camera",
    "edit_camera",
    "upsert_camera",
    "delete_camera",
    "get_camera_controls",
    "update_camera_controls",
    "preview_settings",
    "new_preset",
    "edit_preset",
    "upsert_preset",
    "delete_preset",
    "recall_preset",
    "update_thumbnail",
]



import contextlib
import socket
from dataclasses import dataclass, asdict
from typing import Callable

import requests


@dataclass
class CameraSpec:
    """Info to connect to camera."""

    ip: str
    username: str | None
    password: str | None


@dataclass
class Controls:
    """Camera controls."""

    pan: int
    tilt: int
    zoom: int
    focus: int


@dataclass
class Settings:
    """Camera settings."""

    iris: int | None
    shutter: int | None
    gain: int | None
    drc: int | None
    red_gain: int | None
    blue_gain: int | None
    saturation: int | None
    hue: int | None
    brightness: int | None
    contrast: int | None
    sharpness: int | None
    gamma: int | None
    color_temperature: int | None
    noise2d: int | None
    noise3d: int | None


_sockets: dict[str, socket.socket] = {}


def _get_socket(camera: CameraSpec) -> socket.socket:
    """Return a socket connection to the camera."""
    if camera.ip not in _sockets or _sockets[camera.ip].fileno() == -1:
        _sockets[camera.ip] = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        _sockets[camera.ip].connect((camera.ip, 5678))
        _sockets[camera.ip].settimeout(0.1)
    return _sockets[camera.ip]


def _send_command(camera: CameraSpec, command: str, query: bool = False) -> bytes:
    """Send a command to the camera."""
    print(f"Sending command {command}")
    sock = _get_socket(camera)
    if query:
        with contextlib.suppress(socket.timeout):
            resp = sock.recv(1024)
            while len(resp) == 1024:
                resp = sock.recv(1024)
    preamble = b"\x81" + (b"\x09" if query else b"\x01")
    terminator = b"\xff"
    sock.sendall(preamble + bytearray.fromhex(command) + terminator)
    if query:
        return sock.recv(1024)
    return b""


def _hex_digits(value: int, digits: int) -> str:
    """Return the hex string representation of a value with a fixed number of digits."""
    str_value = f"{value:0{digits}x}"
    return " ".join(f"0{x}" for x in str_value)


def _mod4(value: int) -> int:
    """Return the value modulo 4 hex digits."""
    if value < 0:
        return (value - 1) + 2**16
    return value


# The minimum and maximum values for each command.
_COMMAND_LIMITS: dict[str, tuple[int, int]] = {
    "pan": (-2447, 2448),  # 14.4 steps per degree
    "tilt": (-430, 1296),  # 14.4 steps per degree
    "zoom": (0, 16384),
    "focus": (0, 8192),
    "iris": (0, 12),
    "shutter": (1, 17),
    "gain": (0, 7),
    "drc": (0, 8),
    "red_gain": (0, 255),
    "blue_gain": (0, 255),
    "saturation": (0, 14),
    "hue": (0, 14),
    "brightness": (0, 14),
    "contrast": (0, 14),
    "sharpness": (0, 15),
    "gamma": (0, 4),
    "color_temperature": (0, 55),
    "noise2d": (0, 5),
    "noise3d": (0, 8),
}

# Conversion from commands integers to hex byte strings. Some commands can translate into multiple
#   commands strings if we need to enable a specific setting first. For example, before setting focus,
#   we need to enable manual focus mode.
# See the camera's manual at https://www.adorama.com/col/productManuals/PT30XSDIGYG2.pdf
#   and https://ptzoptics.com/wp-content/uploads/2020/11/PTZOptics-VISCA-over-IP-Rev-1_2-8-20.pdf
_PTZ_FUNCTIONS: dict[str, Callable[[int, int], list[str]]] = {
    "pan_tilt": lambda pan, tilt: [f"06 02 18 14 {_hex_digits(_mod4(pan), 4)} {_hex_digits(_mod4(tilt), 4)}"],
    # Set to manual focus mode, unlock focus, set zoom and focus, and lock focus
    "zoom_focus": lambda zoom, focus: [
        "04 38 03",
        "04 68 03",
        f"04 47 {_hex_digits(zoom, 4)} {_hex_digits(focus, 4)}",
        "04 68 02",
    ],
}
_COMMAND_FUNCTIONS: dict[str, Callable[[int], list[str]]] = {
    # Set manual exposure mode before adjusting these settings
    "iris": lambda value: ["04 39 03", f"04 4B 00 00 {_hex_digits(value, 2)}"],
    "shutter": lambda value: ["04 39 03", f"04 4A 00 00 {_hex_digits(value, 2)}"],
    "gain": lambda value: ["04 39 03", f"04 0C 00 00 {_hex_digits(value, 2)}"],
    "drc": lambda value: ["04 39 03", f"04 25 00 00 {_hex_digits(value, 2)}"],
    # Set manual white balance mode before adjusting these settings
    "red_gain": lambda value: ["04 35 05", f"04 43 00 00 {_hex_digits(value, 2)}"],
    "blue_gain": lambda value: ["04 35 05", f"04 44 00 00 {_hex_digits(value, 2)}"],
    "hue": lambda value: [f"04 4F 00 00 00 {_hex_digits(value, 1)}"],
    "brightness": lambda value: [f"04 A1 00 00 {_hex_digits(value, 2)}"],
    "contrast": lambda value: [f"04 A2 00 00 {_hex_digits(value, 2)}"],
    "sharpness": lambda value: ["04 05 03", f"04 42 00 00 {_hex_digits(value, 2)}"],
    "gamma": lambda value: [f"04 5B {_hex_digits(value, 1)}"],
    "color_temperature": lambda value: [f"04 20 {_hex_digits(value, 2)}"],
    "noise2d": lambda value: ["04 50 03", f"04 53 {_hex_digits(value, 1)}"],
    "noise3d": lambda value: [f"04 54 {_hex_digits(value, 1)}"],
}


def apply_controls(camera: CameraSpec, controls: Controls) -> None:
    """Apply the controls to the camera."""
    for control, value in asdict(controls).items():
        min_value, max_value = _COMMAND_LIMITS[control]
        assert min_value <= value <= max_value, f"{control} value out of range"
    commands = [
        *_PTZ_FUNCTIONS["pan_tilt"](controls.pan, controls.tilt),
        *_PTZ_FUNCTIONS["zoom_focus"](controls.zoom, controls.focus),
    ]
    for command in commands:
        _send_command(camera, command)


def _parse_response(response: bytes) -> int:
    digits = [int(x) for x in response]
    if digits[0] > 8:
        digits[0] -= 16
        digits[-1] += 1
    return sum(digit * 16 ** (len(digits) - i - 1) for i, digit in enumerate(digits))


def read_controls(camera: CameraSpec) -> Controls:
    """Read the current state of controls from the camera."""
    pan_tilt_response = _send_command(camera, "06 12", query=True)
    zoom_response = _send_command(camera, "04 47", query=True)
    focus_response = _send_command(camera, "04 48", query=True)

    for resp in (pan_tilt_response, zoom_response, focus_response):
        assert resp[:2] == b"\x90\x50", f"Invalid response {list(resp)}"

    pan = _parse_response(pan_tilt_response[2:6])
    tilt = _parse_response(pan_tilt_response[6:10])
    zoom = _parse_response(zoom_response[2:6])
    focus = _parse_response(focus_response[2:6])

    return Controls(pan=pan, tilt=tilt, zoom=zoom, focus=focus)


def apply_settings(camera: CameraSpec, settings: Settings) -> None:
    """Apply the settings to the camera."""
    settings_dict: dict[str, int | None] = asdict(settings)
    commands: list[str] = []
    for setting, value in settings_dict.items():
        if value is None:
            continue
        min_value, max_value = _COMMAND_LIMITS[setting]
        assert min_value <= value <= max_value, f"{setting} value out of range"
        if setting in _COMMAND_FUNCTIONS:
            commands.extend(_COMMAND_FUNCTIONS[setting](value))

    for command in commands:
        _send_command(camera, command)

    if settings.saturation is not None:
        requests.get(f"http://{camera.ip}/cgi-bin/param.cgi?post_image_value&saturation/{settings.saturation}")


def fetch_thumbnail(camera: CameraSpec) -> bytes:
    """Fetch the camera thumbnail."""
    requests.get(f"http://{camera.ip}/cgi-bin/snapshot.cgi?post_snapshot_conf&resolution=1920x1080")
    response = requests.get(f"http://{camera.ip}/snapshot.jpg")
    return response.content



import json
from dataclasses import asdict

import dacite
from django.http import Http404, HttpRequest, HttpResponse, JsonResponse
from django.shortcuts import render
from ptz_app.functions.camera import (
    CameraSpec,
    Controls,
    apply_controls,
    read_controls,
    apply_settings,
    Settings,
)

CAMERA_TEMPLATE = "ptz_app/camera.html"


def new_camera(request: HttpRequest) -> HttpResponse:
    """Create a new camera"""
    return render(request, CAMERA_TEMPLATE, {"camera": "null"})


def edit_camera(request: HttpRequest, camera_id: int) -> HttpResponse:
    """Edit a camera"""
    from ptz_app.models import Camera

    camera = Camera.objects.get(id=camera_id)
    if camera is None:
        raise Http404("Camera not found")
    return render(request, CAMERA_TEMPLATE, {"camera": json.dumps(camera.json())})


def upsert_camera(request: HttpRequest) -> JsonResponse:
    """Create or update a camera"""
    from ptz_app.models import Camera, CameraSettings

    assert request.method == "POST"
    body = json.loads(request.body)

    settings = body["default_settings"]
    default_settings = CameraSettings(
        iris=settings.get("iris"),
        shutter=settings.get("shutter"),
        gain=settings.get("gain"),
        drc=settings.get("drc"),
        red_gain=settings.get("red_gain"),
        blue_gain=settings.get("blue_gain"),
        saturation=settings.get("saturation"),
        hue=settings.get("hue"),
        brightness=settings.get("brightness"),
        contrast=settings.get("contrast"),
        sharpness=settings.get("sharpness"),
        gamma=settings.get("gamma"),
        color_temperature=settings.get("color_temperature"),
        noise2d=settings.get("noise2d"),
        noise3d=settings.get("noise3d"),
    )
    if settings.get("id") is not None:
        default_settings.id = settings["id"]
    default_settings.save()

    camera = Camera(
        name=body["name"],
        ip=body["ip"],
        username=body["username"],
        password=body["password"],
        default_settings=default_settings,
    )
    if body.get("id") is not None:
        camera.id = body["id"]
    camera.save()

    return JsonResponse({"status": "ok"})


def delete_camera(request: HttpRequest, camera_id: int) -> JsonResponse:
    """Delete a camera"""
    from ptz_app.models import Camera

    assert request.method == "DELETE"

    camera = Camera.objects.get(id=camera_id)
    if camera is not None:
        camera.delete()
    return JsonResponse({"status": "ok"})


def get_camera_controls(request: HttpRequest, camera_id: int) -> JsonResponse:
    """Get the camera controls."""
    from ptz_app.models import Camera

    camera = Camera.objects.get(id=camera_id)
    if camera is None:
        return JsonResponse({"status": "error", "message": "Camera not found"})

    camera_spec = CameraSpec(ip=camera.ip, username=camera.username, password=camera.password)
    controls = read_controls(camera_spec)

    return JsonResponse(asdict(controls))


def update_camera_controls(request: HttpRequest, camera_id: int) -> JsonResponse:
    """Update the camera controls."""
    from ptz_app.models import Camera

    assert request.method == "POST"
    body = json.loads(request.body)

    camera = Camera.objects.get(id=camera_id)
    if camera is None:
        return JsonResponse({"status": "error", "message": "Camera not found"})

    camera_spec = CameraSpec(ip=camera.ip, username=camera.username, password=camera.password)
    controls = Controls(
        pan=body["pan"],
        tilt=body["tilt"],
        zoom=body["zoom"],
        focus=body["focus"],
    )
    apply_controls(camera_spec, controls)

    return JsonResponse({"status": "ok"})


def preview_settings(request: HttpRequest, settings_id: int) -> JsonResponse:
    """Apply the settings to the camera."""
    from ptz_app.models import Camera, CameraSettings, Preset

    assert request.method == "POST"

    settings = CameraSettings.objects.get(id=settings_id)
    if settings is None:
        return JsonResponse({"status": "error", "message": "Settings not found"})

    preset: Preset | None = settings.preset_set.first()
    camera: Camera | None = settings.camera_set.first()
    if camera is None:
        assert preset is not None
        camera = preset.camera
    assert camera is not None
    camera_spec = CameraSpec(ip=camera.ip, username=camera.username, password=camera.password)

    json_settings = json.loads(request.body)
    if preset is not None:
        for key, value in preset.json().items():
            if json_settings.get(key) is None:
                json_settings[key] = value
    settings_preview = dacite.from_dict(Settings, json_settings)

    apply_settings(camera_spec, settings_preview)

    return JsonResponse({"status": "ok"})



# Generated by Django 5.0.3 on 2024-03-26 17:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("ptz_app", "0002_alter_preset_thumbnail"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="camerasettings",
            name="aperture",
        ),
        migrations.AddField(
            model_name="camerasettings",
            name="blue_gain",
            field=models.IntegerField(null=True),
        ),
        migrations.AddField(
            model_name="camerasettings",
            name="color_temperature",
            field=models.IntegerField(null=True),
        ),
        migrations.AddField(
            model_name="camerasettings",
            name="drc",
            field=models.IntegerField(null=True),
        ),
        migrations.AddField(
            model_name="camerasettings",
            name="gain",
            field=models.IntegerField(null=True),
        ),
        migrations.AddField(
            model_name="camerasettings",
            name="gamma",
            field=models.IntegerField(null=True),
        ),
        migrations.AddField(
            model_name="camerasettings",
            name="iris",
            field=models.IntegerField(null=True),
        ),
        migrations.AddField(
            model_name="camerasettings",
            name="noise2d",
            field=models.IntegerField(null=True),
        ),
        migrations.AddField(
            model_name="camerasettings",
            name="noise3d",
            field=models.IntegerField(null=True),
        ),
        migrations.AddField(
            model_name="camerasettings",
            name="red_gain",
            field=models.IntegerField(null=True),
        ),
        migrations.AddField(
            model_name="camerasettings",
            name="shutter",
            field=models.IntegerField(null=True),
        ),
        migrations.AlterField(
            model_name="preset",
            name="focus",
            field=models.IntegerField(),
        ),
        migrations.AlterField(
            model_name="preset",
            name="pan",
            field=models.IntegerField(),
        ),
        migrations.AlterField(
            model_name="preset",
            name="tilt",
            field=models.IntegerField(),
        ),
        migrations.AlterField(
            model_name="preset",
            name="zoom",
            field=models.IntegerField(),
        ),
    ]



import json
from django.http import Http404
import dacite
from django.http import HttpRequest, HttpResponse, JsonResponse
from django.shortcuts import render

from ptz_app.functions.camera import (
    CameraSpec,
    Controls,
    Settings,
    apply_settings,
    apply_controls,
    fetch_thumbnail,
)

PRESET_TEMPLATE = "ptz_app/preset.html"


def new_preset(request: HttpRequest, camera_id: int) -> HttpResponse:
    """Create a new camera preset."""
    return render(request, PRESET_TEMPLATE, {"preset": "null", "camera_id": camera_id})


def edit_preset(request: HttpRequest, preset_id: int) -> HttpResponse:
    """Edit a camera preset."""
    from ptz_app.models import Preset

    preset = Preset.objects.get(id=preset_id)
    if preset is None:
        raise Http404("Preset not found")
    return render(
        request,
        PRESET_TEMPLATE,
        {"preset": json.dumps(preset.json()), "camera_id": preset.camera.id},
    )


def upsert_preset(request: HttpRequest) -> JsonResponse:
    """Create or update a camera preset."""
    from ptz_app.models import Camera, CameraSettings, Preset

    assert request.method == "POST"
    body = json.loads(request.body)

    camera: Camera = Camera.objects.get(id=body["camera_id"])
    if camera is None:
        return JsonResponse({"status": "error", "message": "Camera not found"})

    settings = body["settings"]
    preset_settings = CameraSettings(
        iris=settings.get("iris"),
        shutter=settings.get("shutter"),
        gain=settings.get("gain"),
        drc=settings.get("drc"),
        red_gain=settings.get("red_gain"),
        blue_gain=settings.get("blue_gain"),
        saturation=settings.get("saturation"),
        hue=settings.get("hue"),
        brightness=settings.get("brightness"),
        contrast=settings.get("contrast"),
        sharpness=settings.get("sharpness"),
        gamma=settings.get("gamma"),
        color_temperature=settings.get("color_temperature"),
        noise2d=settings.get("noise2d"),
        noise3d=settings.get("noise3d"),
    )
    if settings.get("id") is not None:
        preset_settings.id = settings["id"]
    preset_settings.save()

    camera_spec = CameraSpec(ip=camera.ip, username=camera.username, password=camera.password)
    controls = Controls(pan=body["pan"], tilt=body["tilt"], zoom=body["zoom"], focus=body["focus"])
    apply_controls(camera_spec, controls)
    apply_settings(camera_spec, preset_settings.to_dataclass())
    thumbnail = fetch_thumbnail(camera_spec)

    preset = Preset(
        name=body["name"],
        order=body["order"],
        thumbnail=thumbnail,
        camera=camera,
        settings=preset_settings,
        pan=body["pan"],
        tilt=body["tilt"],
        zoom=body["zoom"],
        focus=body["focus"],
    )
    if body.get("id") is not None:
        preset.id = body["id"]
    preset.save()

    return JsonResponse({"status": "ok"})


def delete_preset(request: HttpRequest, preset_id: int) -> JsonResponse:
    """Delete a camera preset."""
    from ptz_app.models import Preset

    assert request.method == "DELETE"

    preset = Preset.objects.get(id=preset_id)
    if preset is None:
        return JsonResponse({"status": "error", "message": "Preset not found"})
    preset.delete()

    return JsonResponse({"status": "ok"})


def recall_preset(request: HttpRequest, preset_id: int) -> JsonResponse:
    """Recall a camera preset."""
    from ptz_app.models import Preset

    preset: Preset = Preset.objects.get(id=preset_id)
    if preset is None:
        return JsonResponse({"status": "error", "message": "Preset not found"})

    camera_spec = CameraSpec(
        ip=preset.camera.ip,
        username=preset.camera.username,
        password=preset.camera.password,
    )
    controls = Controls(pan=preset.pan, tilt=preset.tilt, zoom=preset.zoom, focus=preset.focus)
    apply_controls(camera_spec, controls)

    apply_settings(camera_spec, preset.settings.to_dataclass(preset.camera.default_settings))

    return JsonResponse({"status": "ok"})


def update_thumbnail(request: HttpRequest, preset_id: int) -> JsonResponse:
    """Update the thumbnail for a camera preset."""
    from ptz_app.models import Preset

    assert request.method == "POST"

    preset: Preset = Preset.objects.get(id=preset_id)
    if preset is None:
        return JsonResponse({"status": "error", "message": "Preset not found"})

    body = json.loads(request.body)

    camera_spec = CameraSpec(
        ip=preset.camera.ip,
        username=preset.camera.username,
        password=preset.camera.password,
    )
    if "controls" in body:
        controls = dacite.from_dict(Controls, body["controls"])
    else:
        controls = Controls(pan=preset.pan, tilt=preset.tilt, zoom=preset.zoom, focus=preset.focus)
    apply_controls(camera_spec, controls)

    if "settings" in body:
        settings = dacite.from_dict(Settings, body["settings"])
    else:
        settings = preset.settings.to_dataclass()
    apply_settings(camera_spec, settings)

    preset.thumbnail = fetch_thumbnail(camera_spec)
    preset.save()

    return JsonResponse(preset.json())



"""CLI to run the app."""

import os
import threading
import time
import webbrowser

import click
from django.core.management import execute_from_command_line


def open_browser() -> None:
    """Open the browser."""

    def _open_browser() -> None:
        time.sleep(1)
        webbrowser.open("http://localhost:8000")

    thread = threading.Thread(target=_open_browser, daemon=True)
    thread.start()


@click.command()
@click.option("--no-migrations", is_flag=True, help="Do not run migrations")
@click.option("--no-browser", is_flag=True, help="Do not open the browser")
def main(no_migrations: bool, no_browser: bool) -> None:
    """Run migrations, start the app, and open the browser."""
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "ptz_project.settings")
    if not no_migrations:
        execute_from_command_line(["manage.py", "migrate"])
    if not no_browser:
        open_browser()
    execute_from_command_line(["manage.py", "runserver"])



import json

from django.http import HttpRequest, HttpResponse
from django.shortcuts import render


def index(request: HttpRequest) -> HttpResponse:
    """Render the index page"""
    from ptz_app.models import Camera

    context = {"cameras": json.dumps([camera.json() for camera in Camera.objects.all()])}
    return render(request, "ptz_app/index.html", context)



from django.urls import path

from ptz_app import views

urlpatterns = [
    path("", views.index, name="index"),
    path("cameras/new", views.new_camera, name="new_camera"),
    path("cameras/<int:camera_id>", views.edit_camera, name="edit_camera"),
    path("cameras/upsert", views.upsert_camera, name="upsert_camera"),
    path("cameras/delete/<int:camera_id>", views.delete_camera, name="delete_camera"),
    path("cameras/controls/<int:camera_id>", views.get_camera_controls, name="get_camera_controls"),
    path("cameras/update-controls/<int:camera_id>", views.update_camera_controls, name="update_camera_controls"),
    path("settings/display/<int:settings_id>", views.preview_settings, name="apply_settings"),
    path("presets/new/<int:camera_id>", views.new_preset, name="new_preset"),
    path("presets/<int:preset_id>", views.edit_preset, name="edit_preset"),
    path("presets/upsert", views.upsert_preset, name="upsert_preset"),
    path("presets/delete/<int:preset_id>", views.delete_preset, name="delete_preset"),
    path("presets/recall/<int:preset_id>", views.recall_preset, name="recall_preset"),
    path("presets/update-thumbnail/<int:preset_id>", views.update_thumbnail, name="update_thumbnail"),
]
